// box-counter.c (c) 2016,2022 Claude Heiland-Allen <claude@mathr.co.uk>
// gcc -std=c99 -Wall -Wextra -pedantic -O3 -fopenmp -o box-counter box-counter.c -lm

// feature test for getopt_long
#define _GNU_SOURCE

#include <assert.h>
#include <ctype.h>
#include <inttypes.h>
#include <math.h>
#include <stdint.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <getopt.h>
#include <png.h>
#include <zlib.h>

int64_t max(int64_t a, int64_t b) {
  return a > b ? a : b;
}

int64_t abs64(int64_t x) {
  return x >= 0 ? x : -x;
}

void version(const char *progname) {
  fprintf(stderr,
    "%s 0.2 (c) 2016,2022 Claude Heiland-Allen <claude@mathr.co.uk>\n"
  , progname);
  exit(0);
}

void help(const char *progname) {
  fprintf(stderr,
    "usage: %s [options]\n"
    "where options include:\n"
    "  -? --help              show this help\n"
    "  -V --version           print program version\n"
    "  -i --input in.png      source image (required)\n"
    "  -o --output out.png    output images (out.00.png out.01.png etc)\n"
    "  -e --edge              detect edges\n"
    "  -E --no-edge           don't detect edges (default)\n"
    "  -r --start-radius n    regression start is 2^n pixels (default 1)\n"
    "  -R --end-radius n      regression end is 2^n pixels (default 3)\n"
    "  -m --mask-colour c     background RGB (#rgb or #rrggbb in hex)\n"
    "  -b --box-colour c      draw boxes RGBA (#rgba or #rrggbbaa in hex)\n"
    "  -B --border-colour c   box border RGBA (#rgba or #rrggbbaa in hex)\n"
    "  -v --verbose           output table of box size with counts\n"
  , progname);
  exit(0);
}

struct arguments {
  int verbose;
  int version;
  int help;
  int edge;
  int start_radius;
  int end_radius;
  char *input;
  char *output;
  unsigned char mask_colour[3];
  unsigned char box_colour[4];
  unsigned char border_colour[4];
};

struct arguments args =
{ 0, 0, 0, 0, 1, 3, 0, 0
, { 0xFF, 0xFF, 0xFF } // white
, { 0xFF, 0x00, 0x00, 0x80 } // transparent red
, { 0x80, 0x00, 0x00, 0xFF } // opaque dark red
};

int64_t hex1(int64_t c) {
  if ('0' <= c && c <= '9') { return c - '0'; }
  if ('A' <= c && c <= 'F') { return c - 'A' + 10; }
  if ('a' <= c && c <= 'f') { return c - 'a' + 10; }
  return -1;
}

int64_t hex2(int64_t c1, int64_t c2) {
  return (hex1(c1) << 4) | hex1(c2);
}

void parse_colour(unsigned char *out, const char *hex, int alpha, const char *progname) {
  int i = 0;
  if (! out) { goto fail; }
  if (! hex) { goto fail; }
  if (hex[i++] != '#') { goto fail; }
  if (! isxdigit(hex[i++])) { goto fail; }
  if (! isxdigit(hex[i++])) { goto fail; }
  if (! isxdigit(hex[i++])) { goto fail; }
  if (alpha) {
    if (! isxdigit(hex[i++])) { goto fail; }
  }
  if (isxdigit(hex[i])) {
    if (! isxdigit(hex[i++])) { goto fail; }
    if (! isxdigit(hex[i++])) { goto fail; }
    if (! isxdigit(hex[i++])) { goto fail; }
    if (alpha) {
      if (! isxdigit(hex[i++])) { goto fail; }
    }
  }
  if (hex[i] != 0) { goto fail; }
  --i;
  if (i <= 4) {
    out[0] = hex2(hex[1], hex[1]);
    out[1] = hex2(hex[2], hex[2]);
    out[2] = hex2(hex[3], hex[3]);
    if (alpha) {
      out[3] = hex2(hex[4], hex[4]);
    }
  } else if (i <= 8) {
    out[0] = hex2(hex[1], hex[2]);
    out[1] = hex2(hex[3], hex[4]);
    out[2] = hex2(hex[5], hex[6]);
    if (alpha) {
      out[3] = hex2(hex[7], hex[8]);
    }
  } else {
    goto fail;
  }
  return;
fail:
  fprintf(stderr, "%s: error: bad colour: '%s'\n", progname, hex);
  exit(1);
}

void parse_arguments(int argc, char **argv) {
  int c;
  while (1) {
    int option_index = 0;
    static const char *short_options = "i:o:eEr:R:m:b:B:vV?";
    static struct option long_options[] =
      { { "input",         required_argument, 0, 'i' }
      , { "output",        required_argument, 0, 'o' }
      , { "edge",          no_argument,       0, 'e' }
      , { "no-edge",       no_argument,       0, 'E' }
      , { "start-radius",  required_argument, 0, 'r' }
      , { "end-radius",    required_argument, 0, 'R' }
      , { "mask-colour",   required_argument, 0, 'm' }
      , { "box-colour",    required_argument, 0, 'b' }
      , { "border-colour", required_argument, 0, 'B' }
      , { "verbose",       no_argument,       0, 'v' }
      , { "version",       no_argument,       0, 'V' }
      , { "help",          no_argument,       0, '?' }
      , { 0, 0, 0, 0 }
      };
    c = getopt_long(argc, argv, short_options, long_options, &option_index);
    if (c == -1) { break; }
    switch (c) {
    case 'i': args.input = optarg; break;
    case 'o': args.output = optarg; break;
    case 'e': args.edge = 1; break;
    case 'E': args.edge = 0; break;
    case 'r': args.start_radius = atoi(optarg); break;
    case 'R': args.end_radius = atoi(optarg); break;
    case 'm': parse_colour(&args.mask_colour[0], optarg, 0, argv[0]); break;
    case 'b': parse_colour(&args.box_colour[0], optarg, 1, argv[0]); break;
    case 'B': parse_colour(&args.border_colour[0], optarg, 1, argv[0]); break;
    case 'v': args.verbose = 1; break;
    case 'V': args.version = 1; break;
    case '?': args.help = 1; break;
    default: break;
    }
  }
  if (optind < argc) {
    fprintf(stderr, "%s: error: extra arguments supplied\n", argv[0]);
    exit(1);
  }
  if (args.start_radius >= args.end_radius) {
    fprintf(stderr, "%s: error: start radius must be less than end radius\n", argv[0]);
    exit(1);
  }
}

struct image {
  int64_t width;
  int64_t height;
  int64_t channels;
  unsigned char *data;
};

struct image *alloc_image(int64_t width, int64_t height, int64_t channels) {
  struct image *im = malloc(sizeof(*im));
  im->width = width;
  im->height = height;
  im->channels = channels;
  im->data = calloc(1, (size_t) width * height * channels);
  return im;
}

void free_image(struct image *im) {
  free(im->data);
  free(im);
}

struct image *read_image(const char *filename) {
  FILE *fin = fopen(filename, "rb");
  if (! fin) {
    return 0;
  }
  unsigned char sig[8];
  fread(sig, 1, 8, fin);
  if (! png_check_sig(sig, 8)) {
    fclose(fin);
    return 0;
  }
  png_structp png_ptr = png_create_read_struct(PNG_LIBPNG_VER_STRING, 0, 0, 0);
  if (! png_ptr) {
    fclose(fin);
    return 0;
  }
  png_infop info_ptr = png_create_info_struct(png_ptr);
  if (! info_ptr) {
    fclose(fin);
    png_destroy_read_struct(&png_ptr, 0, 0);
    return 0;
  }
  if (setjmp(png_jmpbuf(png_ptr))) {
    fclose(fin);
    png_destroy_read_struct(&png_ptr, &info_ptr, 0);
    return 0;
  }
  png_init_io(png_ptr, fin);
  png_set_sig_bytes(png_ptr, 8);
  png_read_info(png_ptr, info_ptr);
  png_uint_32 pngwidth = 0, pngheight = 0;
  int bit_depth, color_type;
  png_get_IHDR(png_ptr, info_ptr, &pngwidth, &pngheight, &bit_depth, &color_type, 0, 0, 0);
  int64_t width = pngwidth, height = pngheight;
  if (color_type == PNG_COLOR_TYPE_PALETTE)
      png_set_expand(png_ptr);
  if (color_type == PNG_COLOR_TYPE_GRAY && bit_depth < 8)
      png_set_expand(png_ptr);
  if (png_get_valid(png_ptr, info_ptr, PNG_INFO_tRNS))
      png_set_expand(png_ptr);
  if (bit_depth == 16)
      png_set_strip_16(png_ptr);
  if (color_type == PNG_COLOR_TYPE_GRAY ||
      color_type == PNG_COLOR_TYPE_GRAY_ALPHA)
      png_set_gray_to_rgb(png_ptr);
  png_read_update_info(png_ptr, info_ptr);
  int64_t rowbytes = png_get_rowbytes(png_ptr, info_ptr);
  int64_t channels = png_get_channels(png_ptr, info_ptr);
  unsigned char *image_data = malloc(rowbytes * height);
  if (! image_data) {
    fclose(fin);
    png_destroy_read_struct(&png_ptr, &info_ptr, 0);
    return 0;
  }
  png_bytepp row_pointers = malloc(height*sizeof(png_bytep));
  if (! row_pointers) {
    free(image_data);
    fclose(fin);
    png_destroy_read_struct(&png_ptr, &info_ptr, 0);
    return 0;
  }
  for (int64_t i = 0; i < height; ++i) {
    row_pointers[i] = image_data + (size_t) i * rowbytes;
  }
  png_read_image(png_ptr, row_pointers);
  free(row_pointers);
  png_read_end(png_ptr, 0);
  png_destroy_read_struct(&png_ptr, &info_ptr, 0);
  fclose(fin);
  struct image *in = alloc_image(width, height, 3);
  if (! in) {
    free(image_data);
    return 0;
  }
  for (int64_t j = 0; j < height; ++j) {
    for (int64_t i = 0; i < width; ++i) {
      int64_t ko = 3 * (j * width + i);
      int64_t ki = j * rowbytes + channels * i;
      in->data[ko + 0] = image_data[ki + 0];
      in->data[ko + 1] = image_data[ki + 1];
      in->data[ko + 2] = image_data[ki + 2];
    }
  }
  free(image_data);
  return in;
}

int write_image(const struct image *im, const char *filename) {
  assert(im);
  assert(im->channels == 3);
  png_structp png_ptr = png_create_write_struct(PNG_LIBPNG_VER_STRING, 0, 0, 0);
  if (! png_ptr) {
    return 1;
  }
  png_infop info_ptr = png_create_info_struct(png_ptr);
  if (! info_ptr) {
    png_destroy_write_struct(&png_ptr, 0);
    return 1;
  }
  FILE *fout = fopen(filename, "wb");
  if (! fout) {
    png_destroy_write_struct(&png_ptr, &info_ptr);
    return 1;
  }
  if (setjmp(png_jmpbuf(png_ptr))) {
    png_destroy_write_struct(&png_ptr, &info_ptr);
    fclose(fout);
    return 1;
  }
  png_init_io(png_ptr, fout);
  png_set_compression_level(png_ptr, Z_BEST_COMPRESSION);
  png_set_IHDR(png_ptr, info_ptr, im->width, im->height, 8,
    PNG_COLOR_TYPE_RGB, PNG_INTERLACE_NONE,
    PNG_COMPRESSION_TYPE_DEFAULT, PNG_FILTER_TYPE_DEFAULT);
  png_write_info(png_ptr, info_ptr);
  png_bytepp row_pointers = malloc(im->height*sizeof(png_bytep));
  if (! row_pointers) {
    png_destroy_write_struct(&png_ptr, &info_ptr);
    fclose(fout);
    return 1;
  }
  for (int64_t i = 0; i < im->height; ++i) {
    row_pointers[i] = im->data + i * 3 * im->width;
  }
  png_write_image(png_ptr, row_pointers);
  free(row_pointers);
  png_write_end(png_ptr, 0);
  png_destroy_write_struct(&png_ptr, &info_ptr);
  fclose(fout);
  return 0;
}

struct image *mask_image(const struct image *in, const unsigned char *c) {
  assert(in);
  assert(in->channels == 3);
  struct image *out = alloc_image(in->width, in->height, 1);
  #pragma omp parallel for
  for (int64_t j = 0; j < in->height; ++j) {
    for (int64_t i = 0; i < in->width; ++i) {
      int64_t k = (j * in->width + i) * in->channels;
      out->data[j * in->width + i] =
        ( in->data[k+0] == c[0] &&
          in->data[k+1] == c[1] &&
          in->data[k+2] == c[2] ) * 255;
    }
  }
  return out;
}

struct image *edge_image(const struct image *in) {
  assert(in);
  assert(in->channels == 1);
  struct image *out = alloc_image(in->width, in->height, in->channels);
  #pragma omp parallel for
  for (int64_t j = 0; j < in->height; ++j) {
    for (int64_t i = 0; i < in->width; ++i) {
      if (! in->data[j * in->width + i]) {
        int edge = 0;
        for (int64_t jj = j - 1; jj <= j + 1; ++jj) {
          if (! (0 <= jj && jj < in->height)) {
            continue;
          }
          for (int64_t ii = i - 1; ii <= i + 1; ++ii) {
            if (! (0 <= ii && ii < in->width)) {
              continue;
            }
            if (1 != abs64(jj - j) + abs64(ii - i)) {
              continue;
            }
            edge = edge || in->data[jj * in->width + ii];
          }
        }
        out->data[j * out->width + i] = 255 * ! edge;
      } else {
        out->data[j * out->width + i] = 255;
      }
    }
  }
  return out;
}

int64_t roundup2(int64_t x) {
  int64_t y = 1;  // FIXME infinite loop with huge x
  while (y < x) {
    y <<= 1;
  }
  return y;
}

struct image *pad_image(const struct image *in) {
  assert(in->channels == 1);
  int64_t size = max(roundup2(in->width), roundup2(in->height));
  struct image *out = alloc_image(size, size, 1);
  memset(out->data, 255, size * size);
  for (int64_t j = 0; j < in->height; ++j) {
    for (int64_t i = 0; i < in->width; ++i) {
      int64_t ki = j * in->width + i;
      int64_t ko = j * out->width + i;
      out->data[ko] = in->data[ki];
    }
  }
  return out;
}

struct image *downscale_image(const struct image *in) {
  assert(in->channels == 1);
  struct image *out = alloc_image(in->width >> 1, in->height >> 1, 1);
  #pragma omp parallel for
  for (int64_t j = 0; j < in->height >> 1; ++j) {
    for (int64_t i = 0; i < in->width >> 1; ++i) {
      int px = 0;
      for (int64_t jj = 2 * j; jj < 2 * (j + 1); ++jj) {
        for (int64_t ii = 2 * i; ii < 2 * (i + 1); ++ii) {
          px += in->data[jj * in->width + ii];
        }
      }
      // threshold
      px = px < 4 * 255 ? 0 : 255;
      out->data[j * out->width + i] = px;
    }
  }
  return out;
}

struct image_list {
  struct image_list *next;
  struct image *image;
  int level;
  int64_t count;
};

struct image_list *generate_mipmaps(struct image *in) {
  assert(in->channels == 1);
  assert(in->width == in->height);
  assert(in->width == roundup2(in->width));
  int level = 0;
  int64_t size = in->width;
  struct image_list *mipmaps = malloc(sizeof(*mipmaps));
  mipmaps->image = in;
  mipmaps->level = level++;
  mipmaps->next = 0;
  struct image_list *mipmaps2 = mipmaps;
  while (size > 1) {
    struct image_list *next = malloc(sizeof(*next));
    next->image = downscale_image(mipmaps2->image);
    next->level = level++;
    next->next = 0;
    mipmaps2->next = next;
    mipmaps2 = next;
    size >>= 1;
  }
  return mipmaps;
}

void count_boxes(struct image_list *mipmaps) {
  while (mipmaps) {
    struct image *im = mipmaps->image;
    int64_t count = 0;
    int64_t k = 0;
    for (int64_t j = 0; j < im->height; ++j) {
      for (int64_t i = 0; i < im->width; ++i) {
        count += !im->data[k++];
      }
    }
    mipmaps->count = count;
    mipmaps = mipmaps->next;
  }
}

double simple_linear_regression(const double *ys, int n) {
  double xbar = 0;
  double ybar = 0;
  for (int x = 0; x < n; ++x) {
    xbar += x;
    ybar += ys[x];
  }
  xbar /= n;
  ybar /= n;
  double num = 0;
  double den = 0;
  for (int x = 0; x < n; ++x) {
    double dx = x - xbar;
    double dy = ys[x] - ybar;
    num += dx * dy;
    den += dx * dx;
  }
  return num / den;
}

double compute_dimension(const struct image_list *mipmaps) {
  int number = 0;
  const struct image_list *mipmaps2 = mipmaps;
  while (mipmaps2) {
    number++;
    mipmaps2 = mipmaps2->next;
  }
  double *ys = calloc(number, sizeof(*ys));
  mipmaps2 = mipmaps;
  while (mipmaps2) {
    ys[mipmaps2->level] = log2(mipmaps2->count);
    mipmaps2 = mipmaps2->next;
  }
  int s = args.start_radius;
  int e = args.end_radius - args.start_radius + 1;
  assert(0 <= s && s < number);
  assert(0 <= e && e + s < number);
  double dim = fabs(simple_linear_regression(ys + s, e));
  free(ys);
  return dim;
}

struct image *draw_boxes(struct image *in, struct image *mask, int level) {
  assert(in->channels == 3);
  assert(mask->channels == 1);
  struct image *out = alloc_image(in->width, in->height, 3);
  int64_t factor = 1 << level;
  #pragma omp parallel for
  for (int64_t j = 0; j < in->height; ++j) {
    for (int64_t i = 0; i < in->width; ++i) {
      int ir = in->data[3 * (j * in->width + i) + 0];
      int ig = in->data[3 * (j * in->width + i) + 1];
      int ib = in->data[3 * (j * in->width + i) + 2];
      int mv = mask->data[(j / factor) * mask->width + (i / factor)];
      int border
        =  j % factor == 0 || (j + 1) % factor == 0
        || i % factor == 0 || (i + 1) % factor == 0;
      int r = 0, g = 0, b = 0, a = 0;
      if (! mv) {
        if (! border) {
          r = args.box_colour[0];
          g = args.box_colour[1];
          b = args.box_colour[2];
          a = args.box_colour[3];
        } else {
          r = args.border_colour[0];
          g = args.border_colour[1];
          b = args.border_colour[2];
          a = args.border_colour[3];
        }
      }
      int a1 = 0xFF - a;
      r = (r * a + a1 * ir) / 0xFF;
      g = (g * a + a1 * ig) / 0xFF;
      b = (b * a + a1 * ib) / 0xFF;
      out->data[3 * (j * out->width + i) + 0] = r;
      out->data[3 * (j * out->width + i) + 1] = g;
      out->data[3 * (j * out->width + i) + 2] = b;
    }
  }
  return out;
}

char *output_filename(const char *output, int level) {
  char *filename = malloc(strlen(output) + 100);
  snprintf(filename, 100, "%s.%02d.png", output, level);
  return filename;
}

int main(int argc, char **argv) {
  parse_arguments(argc, argv);
  if (args.help) { help(argv[0]); }
  if (args.version) { version(argv[0]); }
  if (! args.input) {
    fprintf(stderr, "%s: error: no input file given\n", argv[0]);
    exit(1);
  }
  struct image *input = read_image(args.input);
  if (! input) {
    fprintf(stderr, "%s: error: couldn't read input\n", argv[0]);
    exit(1);
  }
  struct image *masked = mask_image(input, &args.mask_colour[0]);
  if (! args.output) {
    free_image(input);
    input = 0;
  }
  if (args.edge) {
    struct image *edge = edge_image(masked);
    free_image(masked);
    masked = edge;
  }
  struct image *padded = pad_image(masked);
  free_image(masked);
  struct image_list *mipmaps = generate_mipmaps(padded);
  count_boxes(mipmaps);
  if (args.verbose) {
    struct image_list *mipmaps2 = mipmaps;
    while (mipmaps2) {
      printf("%d %" PRIi64 "\n", mipmaps2->level, mipmaps2->count);
      mipmaps2 = mipmaps2->next;
    }
  } else {
    double dim = compute_dimension(mipmaps);
    printf("%g\n", dim);
  }
  if (args.output) {
    struct image_list *mipmaps2 = mipmaps;
    while (mipmaps2) {
      struct image *output = draw_boxes(input, mipmaps2->image, mipmaps2->level);
      char *filename = output_filename(args.output, mipmaps2->level);
      int err = write_image(output, filename);
      if (err) {
        fprintf(stderr, "%s: error: failed writing %s\n", argv[0], filename);
        exit(1);
      }
      free(filename);
      free_image(output);
      mipmaps2 = mipmaps2->next;
    }
  }
  return 0;
}
